@@ -5,12 +5,11 @@ use warnings;
 
 ##This module draws heavily from perltoot v0.4 from Tom Christiansen.
 
-require 5.00404;  ##Yes, this is underhanded, but makes support for me easier
-		  ##Not only that, but it's the latest "safe" version of
-		  ##Perl5.  01-03 weren't bug free.
+use 5.006;
+
 use vars (qw($VERSION $Tolerance $Min_samples_number));
 
-$VERSION = '3.0604';
+$VERSION = '3.0609';
 
 $Tolerance = 0.0;
 $Min_samples_number = 4;
@@ -19,7 +18,7 @@ package Statistics::Descriptive::Sparse;
 
 use vars qw($VERSION);
 
-$VERSION = '3.0604';
+$VERSION = '3.0609';
 
 use vars qw(%fields);
 use Carp;
@@ -113,75 +112,76 @@ sub _is_permitted
 }
 
 sub add_data {
-  my $self = shift;  ##Myself
-  my $oldmean;
-  my ($min,$mindex,$max,$maxdex,$sum,$sumsq,$count);
-  my $aref;
+    my $self = shift;  ##Myself
+    my $oldmean;
+    my ($min,$mindex,$max,$maxdex,$sum,$sumsq,$count);
+    my $aref;
 
-  if (ref $_[0] eq 'ARRAY') {
-    $aref = $_[0];
-  }
-  else {
-    $aref = \@_;
-  }
-
-  ##If we were given no data, we do nothing.
-  return 1 if (!@{ $aref });
-
-  ##Take care of appending to an existing data set
-  
-  if (!defined($min = $self->min()))
-  {
-      $min = $aref->[$mindex = 0];
-  }
-  else
-  {
-      $mindex = $self->mindex();
-  }
-
-  if (!defined($max = $self->max()))
-  {
-      $max = $aref->[$maxdex = 0];
-  }
-  else
-  {
-      $maxdex = $self->maxdex();
-  }
-
-  $sum = $self->sum();
-  $sumsq = $self->sumsq();
-  $count = $self->count();
-
-  ##Calculate new mean, sumsq, min and max;
-  foreach ( @{ $aref } ) {
-    $sum += $_;
-    $sumsq += $_**2;
-    $count++;
-    if ($_ >= $max) {
-      $max = $_;
-      $maxdex = $count-1;
+    if (ref $_[0] eq 'ARRAY') {
+	$aref = $_[0];
     }
-    if ($_ <= $min) {
-      $min = $_;
-      $mindex = $count-1;
+    else {
+	$aref = \@_;
     }
-  }
 
-  $self->min($min);
-  $self->mindex($mindex);
-  $self->max($max);
-  $self->maxdex($maxdex);
-  $self->sample_range($max - $min);
-  $self->sum($sum);
-  $self->sumsq($sumsq);
-  $self->mean($sum / $count);
-  $self->count($count);
-  ##indicator the value is not cached.  Variance isn't commonly enough
-  ##used to recompute every single data add.
-  $self->_variance(undef);
-  return 1;
+    ##If we were given no data, we do nothing.
+    return 1 if (!@{ $aref });
+
+    ##Take care of appending to an existing data set
+
+    if (!defined($min = $self->min()))
+    {
+	$min = $aref->[$mindex = 0];
+    }
+    else
+    {
+	$mindex = $self->mindex();
+    }
+
+    if (!defined($max = $self->max()))
+    {
+	$max = $aref->[$maxdex = 0];
+    }
+    else
+    {
+	$maxdex = $self->maxdex();
+    }
+
+    $sum   = $self->sum();
+    $sumsq = $self->sumsq();
+    $count = $self->count();
+
+    ##Calculate new mean, sumsq, min and max;
+    foreach ( @{ $aref } ) {
+	$sum += $_;
+	$sumsq += $_**2;
+	$count++;
+	if ($_ >= $max) {
+	    $max = $_;
+	    $maxdex = $count-1;
+	}
+	if ($_ <= $min) {
+	    $min = $_;
+	    $mindex = $count-1;
+	}
+    }
+
+    $self->min($min);
+    $self->mindex($mindex);
+    $self->max($max);
+    $self->maxdex($maxdex);
+    $self->sample_range($max - $min);
+    $self->sum($sum);
+    $self->sumsq($sumsq);
+    $self->mean($sum / $count);
+    $self->count($count);
+    ##indicator the value is not cached.  Variance isn't commonly enough
+    ##used to recompute every single data add.
+    $self->_variance(undef);
+    return 1;
 }
 
+
 sub standard_deviation {
   my $self = shift;  ##Myself
   return undef if (!$self->count());
@@ -190,35 +190,34 @@ sub standard_deviation {
 
 ##Return variance; if needed, compute and cache it.
 sub variance {
-  my $self = shift;  ##Myself
+    my $self = shift;  ##Myself
 
-  return undef if (!$self->count());
-  
-  my $div = @_ ? 0 : 1;
-  my $count = $self->count();
-  if ($count < 1 + $div) {
-      return 0;
-  }
+    my $count = $self->count();
 
-  if (!defined($self->_variance())) {
-    my $variance = ($self->sumsq()- $count * $self->mean()**2);
+    return undef if !$count;
 
-    # Sometimes due to rounding errors we get a number below 0.
-    # This makes sure this is handled as gracefully as possible.
-    #
-    # See:
-    #
-    # https://rt.cpan.org/Public/Bug/Display.html?id=46026
-    if ($variance < 0)
-    {
-        $variance = 0;
+    return 0 if $count == 1;
+
+    if (!defined($self->_variance())) {
+        my $variance = ($self->sumsq()- $count * $self->mean()**2);
+
+        # Sometimes due to rounding errors we get a number below 0.
+        # This makes sure this is handled as gracefully as possible.
+        #
+        # See:
+        #
+        # https://rt.cpan.org/Public/Bug/Display.html?id=46026
+
+        $variance = $variance < 0 ? 0 : $variance / ($count - 1);
+
+        $self->_variance($variance);
+
+        #  Return now to avoid re-entering this sub
+	#  (and therefore save time when many objects are used).
+        return $variance;
     }
 
-    $variance /= $count - $div;
-
-    $self->_variance($variance);
-  }
-  return $self->_variance();
+    return $self->_variance();
 }
 
 ##Clear a stat.  More efficient than destroying an object and calling
@@ -228,7 +227,7 @@ sub clear {
   my $key;
 
   return if (!$self->count());
-  while (my($field, $value) = each %fields) {
+  while (my($field, $value) = each %fields) {  #  could use a slice assignment here
     $self->{$field} = $value;
   }
 }
@@ -239,7 +238,7 @@ package Statistics::Descriptive::Full;
 
 use vars qw($VERSION);
 
-$VERSION = '3.0604';
+$VERSION = '3.0609';
 
 use Carp;
 use POSIX ();
@@ -249,6 +248,9 @@ use vars qw(@ISA $a $b %fields);
 
 @ISA = qw(Statistics::Descriptive::Sparse);
 
+use List::MoreUtils ();
+use List::Util ();
+
 ##Create a list of fields not to remove when data is updated
 %fields = (
   _permitted => undef,  ##Place holder for the inherited key hash
@@ -261,7 +263,7 @@ use vars qw(@ISA $a $b %fields);
 __PACKAGE__->_make_private_accessors(
     [qw(data samples frequency geometric_mean harmonic_mean
         least_squares_fit median mode
-        skewness kurtosis
+        skewness kurtosis median_absolute_deviation
        )
     ]
 );
@@ -287,7 +289,7 @@ sub new {
   my $proto = shift;
   my $class = ref($proto) || $proto;
   # Create my self re SUPER
-  my $self = $class->SUPER::new();  
+  my $self = $class->SUPER::new();
   bless ($self, $class);  #Re-anneal the object
   $self->_clear_fields();
   return $self;
@@ -305,15 +307,18 @@ sub _delete_all_cached_keys
 {
     my $self = shift;
 
+    my %keys = %{ $self };
+
+    # Remove reserved keys for this class from the deletion list
+    delete @keys{keys %{$self->_reserved}};
+    delete @keys{keys %{$self->_permitted}};
+    delete $keys{_trimmed_mean_cache};
+
     KEYS_LOOP:
-    foreach my $key (keys %{ $self }) { # Check each key in the object
-        # If it's a reserved key for this class, keep it
-        if ($self->_is_reserved($key) || $self->_is_permitted($key))
-        {
-            next KEYS_LOOP;
-        }
-        delete $self->{$key};          # Delete the out of date cached key
+    foreach my $key (keys %keys) { # Check each key in the object
+        delete $self->{$key};  # Delete any out of date cached key
     }
+    $self->{_trimmed_mean_cache} = {};  #  just reset this one
     return;
 }
 
@@ -334,25 +339,76 @@ sub clear {
 }
 
 sub add_data {
-  my $self = shift;
-  my $aref;
+    my $self = shift;  ##Myself
 
-  if (ref $_[0] eq 'ARRAY') {
-    $aref = $_[0];
-  }
-  else {
-    $aref = \@_;
-  }
-  $self->SUPER::add_data($aref);  ##Perform base statistics on the data
-  push @{ $self->_data() }, @{ $aref };
-  ##Clear the presorted flag
-  $self->presorted(0);
+    my $aref;
 
-  $self->_delete_all_cached_keys();
+    if (ref $_[0] eq 'ARRAY') {
+      $aref = $_[0];
+    }
+    else {
+      $aref = \@_;
+    }
 
-  return 1;
+    ##If we were given no data, we do nothing.
+    return 1 if (!@{ $aref });
+
+    my $oldmean;
+    my ($min, $max, $sum, $sumsq);
+    my $count = $self->count;
+
+    #  $count is modified lower down, but we need this flag after that
+    my $has_existing_data = $count;
+
+    # Take care of appending to an existing data set
+    if ($has_existing_data) {
+        $min   = $self->min();
+        $max   = $self->max();
+        $sum   = $self->sum();
+        $sumsq = $self->sumsq();
+    }
+    else {
+        $min   = $aref->[0];
+        $max   = $aref->[0];
+        $sum   = 0;
+        $sumsq = 0;
+    }
+
+    #  need to allow for already having data
+    $sum    += List::Util::sum (@$aref);
+    $sumsq  += List::Util::sum (map {$_ ** 2} @$aref);
+    $max    =  List::Util::max ($max, @$aref);
+    $min    =  List::Util::min ($min, @$aref);
+    $count  +=  scalar @$aref;
+    my $mean = $sum / $count;
+
+    $self->min($min);
+    $self->max($max);
+    $self->sample_range($max - $min);
+    $self->sum($sum);
+    $self->sumsq($sumsq);
+    $self->mean($mean);
+    $self->count($count);
+
+    ##Variance isn't commonly enough
+    ##used to recompute every single data add, so just clear its cache.
+    $self->_variance(undef);
+
+    push @{ $self->_data() }, @{ $aref };
+
+    #  no need to clear keys if we are a newly populated object,
+    #  and profiling shows it takes a long time when creating
+    #  and populating many stats objects
+    if ($has_existing_data) {
+        ##Clear the presorted flag
+        $self->presorted(0);
+        $self->_delete_all_cached_keys();
+    }
+
+    return 1;
 }
 
+
 sub add_data_with_samples {
     my ($self,$aref_values) = @_;
 
@@ -446,6 +502,46 @@ sub get_smoothed_data {
     $self->{_smoother}->get_smoothed_data();
 }
 
+sub maxdex {
+    my $self = shift;
+
+    return undef if !$self->count;
+    my $maxdex;
+
+    if ($self->presorted) {
+        $maxdex = $self->count - 1;
+    }
+    else {
+        my $max = $self->max;
+        $maxdex =  List::MoreUtils::first_index {$_ == $max} $self->get_data;
+    }
+
+    $self->{maxdex} = $maxdex;
+
+    return $maxdex;
+}
+
+sub mindex {
+    my $self = shift;
+
+    return undef if !$self->count;
+    #my $maxdex = $self->{maxdex};
+    #return $maxdex if defined $maxdex;
+    my $mindex;
+
+    if ($self->presorted) {
+        $mindex = 0;
+    }
+    else {
+        my $min = $self->min;
+        $mindex = List::MoreUtils::first_index {$_ == $min} $self->get_data;
+    }
+
+    $self->{mindex} = $mindex;
+
+    return $mindex;
+}
+
 sub sort_data {
   my $self = shift;
 
@@ -454,9 +550,9 @@ sub sort_data {
       ##Sort the data in descending order
       $self->_data([ sort {$a <=> $b} @{$self->_data()} ]);
       $self->presorted(1);
-      ##Fix the maxima and minima indices
-      $self->mindex(0);
-      $self->maxdex($#{$self->_data()});
+      ##Fix the maxima and minima indices - no, this is unnecessary now we have methods
+      #$self->mindex(0);
+      #$self->maxdex($#{$self->_data()});
   }
 
   return 1;
@@ -495,7 +591,7 @@ sub _calc_new_median
 
     ##Even or odd
     if ($count % 2)
-    {   
+    {
         return $self->_data->[($count-1)/2];
     }
     else
@@ -510,8 +606,8 @@ sub _calc_new_median
 sub median {
     my $self = shift;
 
-    return undef if !$self->count;    
-    
+    return undef if !$self->count;
+
     ##Cached?
     if (! defined($self->_median()))
     {
@@ -528,10 +624,10 @@ sub quantile {
        carp("Bad quartile type, must be 0, 1, 2, 3 or 4\n");
        return;
     }
-    
+
     #  check data count after the args are checked - should help debugging
-    return undef if !$self->count;  
-    
+    return undef if !$self->count;
+
     $self->sort_data();
 
     return $self->_data->[0] if ( $QuantileNumber == 0 );
@@ -548,7 +644,7 @@ sub quantile {
     my $aK_quantile     = $self->_data->[ $K_quantile - 1 ];
     return $aK_quantile if ( $F_quantile == 0 );
     my $aKPlus_quantile = $self->_data->[$K_quantile];
-    
+
     # Calcul quantile
     my $quantile = $aK_quantile
       + ( $F_quantile * ( $aKPlus_quantile - $aK_quantile ) );
@@ -562,8 +658,8 @@ sub _real_calc_trimmed_mean
     my $lower = shift;
     my $upper = shift;
 
-    my $lower_trim = int ($self->count()*$lower); 
-    my $upper_trim = int ($self->count()*$upper); 
+    my $lower_trim = int ($self->count()*$lower);
+    my $upper_trim = int ($self->count()*$upper);
     my ($val,$oldmean) = (0,0);
     my ($tm_count,$tm_mean,$index) = (0,0,$lower_trim);
 
@@ -595,7 +691,7 @@ sub trimmed_mean
     }
 
     #  check data count after the args
-    return undef if !$self->count;    
+    return undef if !$self->count;
 
     ##Cache
     my $thistm = join ':',$lower,$upper;
@@ -688,7 +784,7 @@ sub mode
 
 sub geometric_mean {
     my $self = shift;
-    
+
     return undef if !$self->count;
 
     if (!defined($self->_geometric_mean()))
@@ -718,23 +814,22 @@ sub skewness {
     {
         my $n    = $self->count();
         my $sd   = $self->standard_deviation();
-        
+
         my $skew;
-        
+
         #  skip if insufficient records
         if ( $sd && $n > 2) {
-            
+
             my $mean = $self->mean();
-            
+
             my $sum_pow3;
-            
             foreach my $rec ( $self->get_data ) {
-                my $value  = (($rec - $mean) / $sd);
-                $sum_pow3 +=  $value ** 3;
+                $sum_pow3 += (($rec - $mean) / $sd) ** 3;
             }
-            
+
+
             my $correction = $n / ( ($n-1) * ($n-2) );
-            
+
             $skew = $correction * $sum_pow3;
         }
 
@@ -750,25 +845,25 @@ sub kurtosis {
     if (!defined($self->_kurtosis()))
     {
         my $kurt;
-        
+
         my $n  = $self->count();
-        my $sd   = $self->standard_deviation();
-        
+        my $sd = $self->standard_deviation();
+
         if ( $sd && $n > 3) {
 
             my $mean = $self->mean();
-            
+
             my $sum_pow4;
             foreach my $rec ( $self->get_data ) {
                 $sum_pow4 += ( ($rec - $mean ) / $sd ) ** 4;
             }
-            
+
             my $correction1 = ( $n * ($n+1) ) / ( ($n-1) * ($n-2) * ($n-3) );
             my $correction2 = ( 3  * ($n-1) ** 2) / ( ($n-2) * ($n-3) );
-            
+
             $kurt = ( $correction1 * $sum_pow4 ) - $correction2;
         }
-        
+
         $self->_kurtosis($kurt);
     }
 
@@ -914,12 +1009,26 @@ sub least_squares_fit {
   }
 
   $rms = sqrt($rms / $count);
-  
+
   $self->_least_squares_fit([$q, $m, $r, $rms]);
 
   return @{ $self->_least_squares_fit() };
 }
 
+sub median_absolute_deviation {
+    my ($self) = @_;
+
+    if (!defined($self->_median_absolute_deviation()))
+    {
+        my $stat = $self->new;
+        $stat->add_data(map { abs($_ - $self->median) } $self->get_data);
+        $self->_median_absolute_deviation($stat->median);
+    }
+
+    return $self->_median_absolute_deviation();
+}
+
+
 1;
 
 package Statistics::Descriptive;
@@ -982,7 +1091,7 @@ except more efficient.
 
 =item $stat->add_data(1,2,3);
 
-Adds data to the statistics variable. The cached statistical values are 
+Adds data to the statistics variable. The cached statistical values are
 updated automatically.
 
 =item $stat->count();
@@ -1045,7 +1154,7 @@ described above.
 
 Adds data to the statistics variable.  All of the sparse statistical
 values are updated and cached.  Cached values from Full methods are
-deleted since they are no longer valid.  
+deleted since they are no longer valid.
 
 I<Note:  Calling add_data with an empty array will delete all of your
 Full method cached values!  Cached values for the sparse methods are
@@ -1108,7 +1217,7 @@ Example #2: Valid code reference
 
     my $stat = Statistics::Descriptive::Full->new();
     $stat->add_data( 1, 1, 1, 100, 1, );
-  
+
     print $stat->set_outlier_filter( \&outlier_filter ); # => 1
     my @filtered_data = $stat->get_data_without_outliers();
     # @filtered_data is (1, 1, 1, 1)
@@ -1154,9 +1263,9 @@ the flag.
 
 =item $stat->skewness();
 
-Returns the skewness of the data. 
+Returns the skewness of the data.
 A value of zero is no skew, negative is a left skewed tail,
-positive is a right skewed tail. 
+positive is a right skewed tail.
 This is consistent with Excel.
 
 =item $stat->kurtosis();
@@ -1221,7 +1330,7 @@ also return the index of the percentile.
 
 =item $x = $stat->quantile($Type);
 
-Sorts the data and returns estimates of underlying distribution quantiles based on one 
+Sorts the data and returns estimates of underlying distribution quantiles based on one
 or two order statistics from the supplied elements.
 
 This method use the same algorithm as Excel and R language (quantile B<type 7>).
@@ -1236,7 +1345,7 @@ B<$Type> is an integer value between 0 to 4 :
   3 => third quartile (Q3) : upper quartile = highest cut off (25%) of data, or lowest 75% = 75th percentile
   4 => fourth quartile (Q4) : maximal value
 
-Exemple : 
+Exemple :
 
   my @data = (1..10);
   my $stat = Statistics::Descriptive::Full->new();
@@ -1266,12 +1375,12 @@ Returns the geometric mean of the data.
 
 Returns the mode of the data. The mode is the most commonly occuring datum.
 See L<http://en.wikipedia.org/wiki/Mode_%28statistics%29> . If all values
-occur only once, then mode() will return undef. 
+occur only once, then mode() will return undef.
 
 =item $stat->trimmed_mean(ltrim[,utrim]);
 
-C<trimmed_mean(ltrim)> returns the mean with a fraction C<ltrim> 
-of entries at each end dropped. C<trimmed_mean(ltrim,utrim)> 
+C<trimmed_mean(ltrim)> returns the mean with a fraction C<ltrim>
+of entries at each end dropped. C<trimmed_mean(ltrim,utrim)>
 returns the mean after a fraction C<ltrim> has been removed from the
 lower end of the data and a fraction C<utrim> has been removed from the
 upper end of the data.  This method sorts the data before beginning
@@ -1408,7 +1517,7 @@ Previously:
 Colin Kuskie
 
 My email address can be found at http://www.perl.com under Who's Who
-or at: http://search.cpan.org/author/COLINK/.
+or at: https://metacpan.org/author/COLINK .
 
 =head1 CONTRIBUTORS
 
